/* FastXML R package -- very simple yet fast R to libxml2 interface
 *
 * (C)2009 Simon Urbanek   <simon@urbanek.info>
 */

#include <string.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

/* for speed */
#define USE_RINTERNALS 1
#include <Rinternals.h>

static void xmlDocRFinalizer(SEXP sDoc) {
    xmlDocPtr doc = (xmlDocPtr) EXTPTR_PTR(sDoc);
    if (doc) xmlFreeDoc(doc);
}

static int xml_needs_init = 1;

/* the node class is used so often that we really want to cache it */
static SEXP s_fxmlNode, s_children, s_attributes;

static void xml_init() {
    xml_needs_init = 0;
    s_fxmlNode = Rf_mkString("fxmlNode");
    R_PreserveObject(s_fxmlNode);
    s_children = Rf_install("children");
    s_attributes = Rf_install("attributes");
}

SEXP xml_parse(SEXP sFn) {
    const char *fn;
    xmlDocPtr doc;
    SEXP res;
    
    if (TYPEOF(sFn) != STRSXP || LENGTH(sFn) != 1)
	Rf_error("invalid file name");
    
    fn = CHAR(STRING_ELT(sFn, 0));
 
    if (xml_needs_init) xml_init();
    
    doc = xmlReadFile(fn, NULL, 0);
    if (!doc)
	Rf_error("failed to parse %s\n", fn);
    
    res = R_MakeExternalPtr(doc, R_NilValue, R_NilValue);
    R_RegisterCFinalizer(res, xmlDocRFinalizer);
    Rf_setAttrib(res, R_ClassSymbol, Rf_mkString("fxmlDoc"));
    return res;
}

static SEXP node2SEXP(xmlNode *n) {
    SEXP res;
    if (n->_private) return n->_private;
    
    if (n->type == XML_TEXT_NODE && n->content) 
	res = mkString((const char*)n->content);
    else {
	res = R_MakeExternalPtr(n, R_NilValue, R_NilValue);
	Rf_setAttrib(res, R_ClassSymbol, s_fxmlNode);
    }
    n->_private = res;
    return res;
}

static SEXP attr2SEXP(xmlAttr *n) {
    SEXP res;
    if (n->_private) return n->_private;

    /* normally there should be just one text child so we can return directly the content */
    if (n->children && !n->children->next && n->children->content && n->children->type == XML_TEXT_NODE)
	res = mkString((const char*)n->children->content);
    else {
	res = R_MakeExternalPtr(n, R_NilValue, R_NilValue);
	Rf_setAttrib(res, R_ClassSymbol, Rf_mkString("fxmlAttr"));
    }
    n->_private = res;
    return res;
}

SEXP xml_root(SEXP sDoc) {
    xmlDocPtr doc;
    
    if (TYPEOF(sDoc) != EXTPTRSXP)
	Rf_error("invalid document");
    doc = (xmlDocPtr) EXTPTR_PTR(sDoc);
    if (!doc)
	Rf_error("NULL document");
    return node2SEXP(xmlDocGetRootElement(doc));
}

SEXP xml_children(SEXP sNode) {
    xmlNode *node, *cn;
    unsigned int i = 0, n = 0;
    SEXP nam, c = Rf_getAttrib(sNode, s_children);
    if (c != R_NilValue) return c;
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    if (!node)
	Rf_error("NULL node");
    if (!(cn = node->children)) return R_NilValue;
    while (cn) { if (cn->type == XML_ELEMENT_NODE) n++; cn = cn->next; }
    /* Rprintf("xml_children: %d element nodes\n", n); */
    c = Rf_allocVector(VECSXP, n);
    Rf_setAttrib(sNode, s_children, c);
    nam = Rf_allocVector(STRSXP, n);
    Rf_setAttrib(c, R_NamesSymbol, nam);
    cn = node->children;
    while (cn) {
	if (cn->type == XML_ELEMENT_NODE) {
	    SET_VECTOR_ELT(c, i, node2SEXP(cn));
	    if (cn->name) SET_STRING_ELT(nam, i, mkChar((const char*)cn->name));
	    i++;
	}
	cn = cn->next;
    }
    return c;
}

SEXP xml_attributes(SEXP sNode) {
    xmlNode *node;
    xmlAttr *a;
    unsigned int i = 0, n = 0;
    SEXP nam, c = Rf_getAttrib(sNode, s_attributes);
    if (c != R_NilValue) return c;
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    if (!node)
	Rf_error("NULL node");
    if (!(a = node->properties)) return R_NilValue;
    while (a) { n++; a = a->next; }
#if MIXED_XML_ATTR
    c = Rf_allocVector(VECSXP, n);
    Rf_setAttrib(sNode, s_attributes, c);
    nam = Rf_allocVector(STRSXP, n);
    Rf_setAttrib(c, R_NamesSymbol, nam);
    a = node->properties;
    while (a) {
	SET_VECTOR_ELT(c, i, attr2SEXP(a));
	if (a->name) SET_STRING_ELT(nam, i, mkChar((const char*)a->name));
	i++;
	a = a->next;
    }
#else
    c = Rf_allocVector(STRSXP, n);
    Rf_setAttrib(sNode, s_attributes, c);
    nam = Rf_allocVector(STRSXP, n);
    Rf_setAttrib(c, R_NamesSymbol, nam);
    a = node->properties;
    while (a) {
	if (a->children && a->children->content)
	    SET_STRING_ELT(c, i, mkChar((const char *) a->children->content));
	if (a->name)
	    SET_STRING_ELT(nam, i, mkChar((const char*)a->name));
	i++;
	a = a->next;
    }
#endif
    return c;
}

SEXP xml_attr(SEXP sNode, SEXP sName) {
    xmlNode *node;
    xmlAttr *a;
    const char *name;
    
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    a = node->properties;
    if (TYPEOF(sName) != STRSXP || LENGTH(sName) != 1)
	Rf_error("invalid attribute name");
    name = CHAR(STRING_ELT(sName, 0));
    while (a) {
	if (a->name && !strcmp((const char*)a->name, name)) break;
	a = a->next;
    }
    if (!a)
	return R_NilValue;
    return attr2SEXP(a);
}

SEXP xml_attr_names(SEXP sNode) {
    xmlNode *node;
    xmlAttr *a;
    unsigned int i = 0;
    SEXP ns = Rf_getAttrib(sNode, install("attr.names"));
    if (ns != R_NilValue) return ns;
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    a = node->properties;
    while (a) { i++; a = a->next; }
    ns = Rf_allocVector(STRSXP, i);
    Rf_setAttrib(sNode, install("attr.names"), ns);
    i = 0;
    a = node->properties;
    while (a) {
	SET_STRING_ELT(ns, i++, mkChar((const char*)a->name));
	a = a->next;
    }
    return ns;
}

SEXP xml_child(SEXP sNode, SEXP sName) {
    xmlNode *node, *a;
    const char *name;
    
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    a = node->children;
    if (TYPEOF(sName) != STRSXP || LENGTH(sName) != 1)
	Rf_error("invalid child name");
    name = CHAR(STRING_ELT(sName, 0));
    while (a) {
	if (a->name && !strcmp((const char*)a->name, name)) break;
	a = a->next;
    }
    if (!a)
	return R_NilValue;
    return node2SEXP(a);
}

SEXP xml_children_names(SEXP sNode) {
    xmlNode *node, *a;
    unsigned int i = 0;
    SEXP ns = Rf_getAttrib(sNode, install("ch.names"));
    if (ns != R_NilValue) return ns;
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    a = node->children;
    while (a) { i++; a = a->next; }
    ns = Rf_allocVector(STRSXP, i);
    Rf_setAttrib(sNode, install("ch.names"), ns);
    i = 0;
    a = node->children;
    while (a) {
	SET_STRING_ELT(ns, i++, mkChar((const char*)a->name));
	a = a->next;
    }
    return ns;
}

SEXP xml_node_name(SEXP sNode) {
    xmlNode *node;
    SEXP name = Rf_getAttrib(sNode, install("name"));
    if (name != R_NilValue) return name;
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    name = node->name ? mkString((const char*)node->name) : R_NilValue;
    Rf_setAttrib(sNode, install("name"), name);
    return name;
}

SEXP xml_node_content(SEXP sNode) {
    xmlNode *node;
    SEXP content = Rf_getAttrib(sNode, install("content"));
    if (content != R_NilValue) return content;
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    if (node->content) {
	content = mkString((const char*)node->content);
	Rf_setAttrib(sNode, install("content"), content);
    }
    return content;
}

SEXP xml_node_info(SEXP sNode) {
    unsigned int i = 0;
    SEXP res, na;
    xmlNode *node, *n;
    xmlAttr *a;
    if (TYPEOF(sNode) != EXTPTRSXP)
	Rf_error("invalid node");
    node = (xmlNode*) EXTPTR_PTR(sNode);
    PROTECT(res = Rf_allocVector(VECSXP, 4));
    SET_VECTOR_ELT(res, 0, ScalarInteger((int)node->type));
    SET_VECTOR_ELT(res, 1, mkString((const char*)node->name));
    n = node->children; while (n) { i++; n = n->next; }
    SET_VECTOR_ELT(res, 2, ScalarInteger(i));
    i = 0;
    a = node->properties; while (a) { i++; a = a->next; }
    SET_VECTOR_ELT(res, 3, ScalarInteger(i));
    na = Rf_allocVector(STRSXP, 4);
    Rf_setAttrib(res, R_NamesSymbol, na);
    SET_STRING_ELT(na, 0, mkChar("type"));
    SET_STRING_ELT(na, 1, mkChar("name"));
    SET_STRING_ELT(na, 2, mkChar("children"));
    SET_STRING_ELT(na, 3, mkChar("attributes"));
    UNPROTECT(1);
    return res;
}

/* -- ok this has really nothing to do with XML but it is very common
      that you need RSO when parsing XML files so I have included it
      here. Cleary it needs R with RSO support. */
SEXP setRS(SEXP x, SEXP val) {
#ifdef SET_RS_OBJECT
  if (asLogical(val)) SET_RS_OBJECT(x); else UNSET_RS_OBJECT(x);
#endif
  return x;
}

SEXP getRS(SEXP x) {
#ifdef SET_RS_OBJECT
  return ScalarLogical(IS_RS_OBJECT(x) != 0);
#else
  return ScalarLogical(0);
#endif
}
